@@ -916,7 +916,7 @@ func (p *printer) printParameterList(list []*Field, tok token) {
 			}
 			p.print(blank)
 		}
-		p.printNode(Unparen(f.Type)) // no need for (extra) parentheses around parameter types
+		p.printNode(f.Type)
 	}
 	// A type parameter list [P T] where the name P and the type expression T syntactically
 	// combine to another valid (value) expression requires a trailing comma, as in [P *T,]
@@ -931,7 +931,7 @@ func (p *printer) printParameterList(list []*Field, tok token) {
 // combinesWithName reports whether a name followed by the expression x
 // syntactically combines to another valid (value) expression. For instance
 // using *T for x, "name *T" syntactically appears as the expression x*T.
-// On the other hand, using  P|Q or *P|~Q for x, "name P|Q" or name *P|~Q"
+// On the other hand, using  P|Q or *P|~Q for x, "name P|Q" or "name *P|~Q"
 // cannot be combined into a valid (value) expression.
 func combinesWithName(x Expr) bool {
 	switch x := x.(type) {
@@ -943,9 +943,13 @@ func combinesWithName(x Expr) bool {
 		// binary expressions
 		return combinesWithName(x.X) && !isTypeElem(x.Y)
 	case *ParenExpr:
-		// name(x) combines but we are making sure at
-		// the call site that x is never parenthesized.
-		panic("unexpected parenthesized expression")
+		// Note that the parser strips parentheses in these cases
+		// (see extractName, parser.typeOrNil) unless keep_parens
+		// is set, so we should never reach here.
+		// Do the right thing (rather than panic) for testing and
+		// in case we change parser behavior.
+		// See also go.dev/issues/69206.
+		return !isTypeElem(x.X)
 	}
 	return false
 }
