@@ -375,7 +375,7 @@ func traceAdvance(stopTrace bool) {
 			me := getg().m.curg
 			// We don't have to handle this G status transition because we
 			// already eliminated ourselves from consideration above.
-			casGToWaitingForGC(me, _Grunning, waitReasonTraceGoroutineStatus)
+			casGToWaitingForSuspendG(me, _Grunning, waitReasonTraceGoroutineStatus)
 			// We need to suspend and take ownership of the G to safely read its
 			// goid. Note that we can't actually emit the event at this point
 			// because we might stop the G in a window where it's unsafe to write
@@ -524,10 +524,11 @@ func traceAdvance(stopTrace bool) {
 				// trace.lock needed for traceBufFlush, but also to synchronize
 				// with traceThreadDestroy, which flushes both buffers unconditionally.
 				lock(&trace.lock)
-				bufp := &mp.trace.buf[gen%2]
-				if *bufp != nil {
-					traceBufFlush(*bufp, gen)
-					*bufp = nil
+				for exp, buf := range mp.trace.buf[gen%2] {
+					if buf != nil {
+						traceBufFlush(buf, gen)
+						mp.trace.buf[gen%2][exp] = nil
+					}
 				}
 				unlock(&trace.lock)
 
